"""Created on Sat Feb 15 22:51:02 2020.

@author: mofarrag
"""
import datetime as dt
import os
import zipfile
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from loguru import logger
from osgeo import gdal
from scipy.stats import gumbel_r  # genextreme,
from statista.distributions import GEV, Gumbel, PlottingPosition

from Hapi.hm.river import River

# from matplotlib import gridspec


class Inputs(River):
    """Inputs.

    Methods
        1- ExtractHydrologicalInputs
        2- StatisticalProperties
        3- WriteHQFile
        4- ReturnPeriod
        5- ReadRIMResult
        6- CreateTraceALL
    """

    def __init__(self, Name, version=3):
        """Inputs.

            Inputs is an object to create the inputs for the river model

        Parameters
        ----------
        Name: [str]

        version: [int]
            model version
        """
        self.Name = Name
        self.version = version
        self.StatisticalPr = None
        self.DistributionPr = None

    def ExtractHydrologicalInputs(
        self, WeatherGenerator, FilePrefix, realization, path, SWIMNodes, SavePath
    ):
        """ExtractHydrologicalInputs.

        Parameters
        ----------
        WeatherGenerator : TYPE
            DESCRIPTION.
        FilePrefix : TYPE
            DESCRIPTION.
        realization : [Integer]
            type the number of the realization (the order of the 100 year
            run by swim).
        path : [String]
             SWIMResultFile is the naming format you used in naming the result
             files of the discharge values stored with the name of the file as
             out+realization number + .dat (ex out15.dat).
        SWIMNodes : [String]
            text file containing the list of sub-basins IDs or computational nodes ID you
            have used to run SWIM and store the results.
        SavePath : [String]
            path to the folder where you want to save the separate file for
            each sub-basin.

        Returns
        -------
        None.
        """
        if WeatherGenerator:
            """WeatherGenerator."""
            SWIMResultFile = FilePrefix + str(realization) + ".dat"
            # 4-5
            # check whether the the name of the realization the same as the name of 3
            # the saving file or not to prevent any confusion in saving the files
            if int(realization) <= 9:
                assert int(SWIMResultFile[-5:-4]) == int(
                    SavePath[-1]
                ), " Wrong files sync "
            else:
                assert int(SWIMResultFile[-6:-4]) == int(
                    SavePath[-2:]
                ), " Wrong files sync "
        else:
            """Observed data."""
            SWIMResultFile = FilePrefix + str(realization) + ".dat"

        """
        SWIM writes the year as the first colmun then day as a second column and the
        discharge values starts from the thirst column so you have to write number of
        columns to be ignored at the begining
        """
        ignoreColumns = 2

        # read SWIM result file
        SWIMData = pd.read_csv(
            path + "/" + SWIMResultFile, delimiter=r"\s+", header=None
        )
        Nodes = pd.read_csv(path + "/" + SWIMNodes, header=None)

        for i in range(len(Nodes)):
            SWIMData.loc[:, i + ignoreColumns].to_csv(
                SavePath + "/" + str(Nodes.loc[i, 0]) + ".txt", header=None, index=None
            )

    def statisticalProperties(
        self,
        ComputationalNodes: Union[str, list],
        TSdirectory: str,
        start: str,
        WarmUpPeriod: int,
        SavePlots: bool,
        SavePath: str,
        Filter: Union[bool, float, int] = False,
        Distibution: str = "GEV",
        method: str = "lmoments",
        EstimateParameters: bool = False,
        Quartile: float = 0,
        SignificanceLevel: float = 0.1,
        file_extension: str = ".txt",
        date_format: str = "%Y-%m-%d",
    ):
        """StatisticalProperties.

        StatisticalProperties method reads the discharge hydrographs of rainfall
        runoff model and hydraulic model for some computational nodes and
        calculate some statistical properties.

        the code assumes that the time series are of a daily temporal resolution,
        and that the hydrological year is 1-Nov/31-Oct
        (Petrow and Merz, 2009, JoH).

        Parameters
        ----------
        ComputationalNodes : [list]
            The file/list which contains the ID of the computational
            nodes you want to do the statistical analysis for, the ObservedFile
            should contain the discharge time series of these nodes in order.
        TSdirectory : [str]
            The directory where the time series files exist.
        start : [string]
            the begining date of the time series.
        WarmUpPeriod : [integer]
            the number of days you want to neglect at the begining of the
            Simulation (warm up period).
        SavePlots : [Bool]
            True if you want to save the plots.
        SavePath : [String]
            the path where you want to  save the statistical properties.
        Filter: [Bool]
            for observed or RIMresult data it has gaps of times where the
            model did not run or gaps in the observed data if these gap days
            are filled with a specific value and you want to ignore it here
            give Filter = Value you want
        Distibution: [str]
            Default is "GEV".
        method: [str]
            available methods are 'mle', 'mm', 'lmoments', optimization. Default is "lmoments"
        EstimateParameters: [bool]
            Default is False.
        Quartile: [float]
            Default is 0.
        SignificanceLevel:
            Default is [0.1].
        file_extension: [str]
            Default is '.txt'.
        date_format: [str]
            Default is "%Y-%m-%d".

        Returns
        -------
        Statistical Properties.csv:
            file containing some statistical properties like mean, std, min, 5%, 25%,
            median, 75%, 95%, max, t_beg, t_end, nyr, q1.5, q2, q5, q10, q25, q50,
            q100, q200, q500.
        """
        if isinstance(ComputationalNodes, str):
            ComputationalNodes = np.loadtxt(ComputationalNodes, dtype=np.uint16)
        elif isinstance(ComputationalNodes, list):
            ComputationalNodes = ComputationalNodes
        else:
            raise TypeError("ComputationalNodes should be either a path or a list")

        # hydrographs

        TS = pd.DataFrame()
        # for the hydraulic model results
        logger.info(
            "The function ignores the date column in the time series files and starts from the "
            "given start parameter to the function so check if it is the same start date as in "
            "the files"
        )

        skip = []
        for i in range(len(ComputationalNodes)):
            try:
                TS.loc[:, int(ComputationalNodes[i])] = pd.read_csv(
                    f"{TSdirectory}/{int(ComputationalNodes[i])}{file_extension}",
                    skiprows=1,
                    header=None,
                )[1].tolist()
            except FileNotFoundError:
                logger.warning(
                    f" File {int(ComputationalNodes[i])}{file_extension} does not exist"
                )
                skip.append(int(ComputationalNodes[i]))

        StartDate = dt.datetime.strptime(start, date_format)
        EndDate = StartDate + dt.timedelta(days=TS.shape[0] - 1)
        ind = pd.date_range(StartDate, EndDate)
        TS.index = ind

        # neglect the first year (warmup year) in the time series
        TS = TS.loc[StartDate + dt.timedelta(days=WarmUpPeriod) : EndDate, :]

        # List of the table output, including some general data and the return periods.
        col_csv = [
            "mean",
            "std",
            "min",
            "5%",
            "25%",
            "median",
            "75%",
            "95%",
            "max",
            "t_beg",
            "t_end",
            "nyr",
        ]
        rp_name = [
            "q1.5",
            "q2",
            "q5",
            "q10",
            "q25",
            "q50",
            "q100",
            "q200",
            "q500",
            "q1000",
        ]
        col_csv = col_csv + rp_name

        # In a table where duplicates are removed (np.unique), find the number of
        # gauges contained in the .csv file.
        # no_gauge = len(ComputationalNodes)
        # Declare a dataframe for the output file, with as index the gaugne numbers
        # and as columns all the output names.
        StatisticalPr = pd.DataFrame(np.nan, index=ComputationalNodes, columns=col_csv)
        StatisticalPr.index.name = "id"
        if Distibution == "GEV":
            DistributionPr = pd.DataFrame(
                np.nan,
                index=ComputationalNodes,
                columns=["c", "loc", "scale", "D-static", "P-Value"],
            )
        else:
            DistributionPr = pd.DataFrame(
                np.nan,
                index=ComputationalNodes,
                columns=["loc", "scale", "D-static", "P-Value"],
            )
        DistributionPr.index.name = "id"
        # required return periods
        T = [1.5, 2, 5, 10, 25, 50, 50, 100, 200, 500, 1000]
        T = np.array(T)
        # these values are the Non Exceedance probability (F) of the chosen
        # return periods F = 1 - (1/T)
        # Non Exceedance propabilities
        # F = [1/3, 0.5, 0.8, 0.9, 0.96, 0.98, 0.99, 0.995, 0.998]
        F = 1 - (1 / T)
        # Iteration over all the gauge numbers.
        if SavePlots:
            rpath = os.path.join(SavePath, "Figures")
            if not os.path.exists(rpath):
                os.mkdir(rpath)

        for i in ComputationalNodes:
            # if the node file does not exit and was not read in the top of the function
            if i in skip:
                continue

            QTS = TS.loc[:, i]
            # The time series is resampled to the annual maxima, and turned into a
            # numpy array.
            # The hydrological year is 1-Nov/31-Oct (from Petrow and Merz, 2009, JoH).
            amax = QTS.resample("A-OCT").max().values

            if type(Filter) != bool:
                amax = amax[amax != Filter]

            if EstimateParameters:
                # TODO: still to be tested and prepared for GEV
                # estimate the parameters through an optimization
                # alpha = (np.sqrt(6) / np.pi) * amax.std()
                # beta = amax.mean() - 0.5772 * alpha
                # param_dist = [beta, alpha]
                threshold = np.quantile(amax, Quartile)
                if Distibution == "GEV":
                    dist = GEV(amax)
                    param_dist = dist.estimateParameter(
                        method="optimization",
                        ObjFunc=Gumbel.ObjectiveFn,
                        threshold=threshold,
                    )
                else:
                    dist = Gumbel(amax)
                    param_dist = dist.estimateParameter(
                        method="optimization",
                        ObjFunc=Gumbel.ObjectiveFn,
                        threshold=threshold,
                    )
            else:
                # estimate the parameters through an maximum liklehood method
                if Distibution == "GEV":
                    dist = GEV(amax)
                    # defult parameter estimation method is maximum liklihood method
                    param_dist = dist.estimateParameter(method=method)
                else:
                    # A gumbel distribution is fitted to the annual maxima
                    dist = Gumbel(amax)
                    # defult parameter estimation method is maximum liklihood method
                    param_dist = dist.estimateParameter(method=method)

            (
                DistributionPr.loc[i, "D-static"],
                DistributionPr.loc[i, "P-Value"],
            ) = dist.ks()
            if Distibution == "GEV":
                DistributionPr.loc[i, "c"] = param_dist[0]
                DistributionPr.loc[i, "loc"] = param_dist[1]
                DistributionPr.loc[i, "scale"] = param_dist[2]
            else:
                DistributionPr.loc[i, "loc"] = param_dist[0]
                DistributionPr.loc[i, "scale"] = param_dist[1]

            # Return periods from the fitted distribution are stored.
            # get the Discharge coresponding to the return periods
            if Distibution == "GEV":
                Qrp = dist.theporeticalEstimate(
                    param_dist[0], param_dist[1], param_dist[2], F
                )
            else:
                Qrp = dist.theporeticalEstimate(param_dist[0], param_dist[1], F)

            # to get the Non Exceedance probability for a specific Value
            # sort the amax
            amax.sort()
            # calculate the F (Exceedence probability based on weibul)
            cdf_Weibul = PlottingPosition.weibul(amax)
            # Gumbel.probapilityPlot method calculates the theoretical values
            # based on the Gumbel distribution
            # parameters, theoretical cdf (or weibul), and calculate the confidence interval
            if SavePlots:

                if Distibution == "GEV":
                    fig, ax = dist.probapilityPlot(
                        param_dist[0],
                        param_dist[1],
                        param_dist[2],
                        cdf_Weibul,
                        alpha=SignificanceLevel,
                    )
                else:
                    fig, ax = dist.probapilityPlot(
                        param_dist[0],
                        param_dist[1],
                        cdf_Weibul,
                        alpha=SignificanceLevel,
                    )

                fig[0].savefig(f"{SavePath}/Figures/{i}.png", format="png")
                plt.close()

                fig[1].savefig(f"{SavePath}/Figures/F-{i}.png", format="png")
                plt.close()

            StatisticalPr.loc[i, "mean"] = QTS.mean()
            StatisticalPr.loc[i, "std"] = QTS.std()
            StatisticalPr.loc[i, "min"] = QTS.min()
            StatisticalPr.loc[i, "5%"] = QTS.quantile(0.05)
            StatisticalPr.loc[i, "25%"] = QTS.quantile(0.25)
            StatisticalPr.loc[i, "median"] = QTS.quantile(0.50)
            StatisticalPr.loc[i, "75%"] = QTS.quantile(0.75)
            StatisticalPr.loc[i, "95%"] = QTS.quantile(0.95)
            StatisticalPr.loc[i, "max"] = QTS.max()
            StatisticalPr.loc[i, "t_beg"] = QTS.index.min()
            StatisticalPr.loc[i, "t_end"] = QTS.index.max()
            StatisticalPr.loc[i, "nyr"] = (
                StatisticalPr.loc[i, "t_end"] - StatisticalPr.loc[i, "t_beg"]
            ).days / 365.25
            for irp, irp_name in zip(Qrp, rp_name):
                StatisticalPr.loc[i, irp_name] = irp

            # Print for prompt and check progress.
            logger.info(f"Gauge {i} done.")

        # Output file
        StatisticalPr.to_csv(
            f"{SavePath}/Statistical Properties.csv", float_format="%.4f"
        )
        self.StatisticalPr = StatisticalPr
        DistributionPr.to_csv(
            f"{SavePath}/DistributionProperties.csv", float_format="%.4f"
        )
        self.DistributionPr = DistributionPr

    def WriteHQFile(self, NoNodes: int, StatisticalPropertiesFile: str, SaveTo: str):
        """WriteHQFile.

        Parameters
        ----------
        NoNodes:
        StatisticalPropertiesFile:
        SaveTo:
        Returns
        -------
        None
        """
        # Create a table of all nodes and sub-basins
        HQ = pd.DataFrame(
            index=list(range(NoNodes)), columns=["id", "2yrs", "10yrs", "100yrs"]
        )
        HQ.loc[:, ["2yrs", "10yrs", "100yrs"]] = -1
        HQ.loc[:, "id"] = range(1, NoNodes + 1)
        StatisticalPr = pd.read_csv(StatisticalPropertiesFile)
        for i in range(len(StatisticalPr)):
            # i=0
            HQ.loc[StatisticalPr.loc[i, "id"] - 1, "2yrs"] = StatisticalPr.loc[i, "q2"]
            HQ.loc[StatisticalPr.loc[i, "id"] - 1, "10yrs"] = StatisticalPr.loc[
                i, "q10"
            ]
            HQ.loc[StatisticalPr.loc[i, "id"] - 1, "100yrs"] = StatisticalPr.loc[
                i, "q100"
            ]

        # save the HQ file
        HQ.to_csv(SaveTo, float_format="%%.2f".format, index=False, header=None)

    @staticmethod
    def StringSpace(Inp):
        return str(Inp) + "  "

    def returnPeriod(
        self,
        MapsPath,
        prefix,
        DistributionPrF,
        TraceF,
        SubsF,
        replacementF,
        HydrologicalInputsPath,
        SubIDMapF,
        ExtraSubsF,
        Fromfile,
        Tofile,
        SaveTo,
        wpath,
    ):
        AllResults = os.listdir(MapsPath)
        # list of the Max Depth files only
        MaxDepthList = list()
        for i in range(len(AllResults)):
            if AllResults[i].startswith(prefix):
                MaxDepthList.append(AllResults[i])
        # Read Inputs
        # read the Distribution parameters for each upstream computatiopnal node
        DistributionPr = pd.read_csv(DistributionPrF)
        USnode = pd.read_csv(TraceF, header=None)
        USnode.columns = ["SubID", "US", "DS"]
        # get the sub basin Id from the guide file it is the same shape in RIM1.0 and RIM2.0
        SubsID = pd.read_csv(SubsF, header=None, usecols=[0])

        ReplacementSub = pd.read_csv(replacementF)

        # read the hydrograph for all the US nodes
        # StartDate = "1950-1-1"
        # StartDate = dt.datetime.strptime(StartDate,"%Y-%m-%d")
        # ind = pd.date_range(StartDate, StartDate + dt.timedelta(days = NoYears*365), freq = "D")

        ind = range(
            1,
            len(
                pd.read_csv(
                    HydrologicalInputsPath
                    + "/"
                    + str(int(USnode.loc[SubsID.loc[10, 0] - 1, "US"]))
                    + ".txt"
                ).values
            ),
        )

        Hydrographs = pd.DataFrame(index=ind, columns=SubsID[0].to_list())

        for i in range(len(SubsID)):
            #    i=1
            # search for the SubId in the USnode or it is listed by order so subID=343 exist
            # in the row 342 (SubID-1)
            # np.where(USnode['SubID'] == SubsID.loc[i,0])
            try:
                if int(USnode.loc[SubsID.loc[i, 0] - 1, "US"]) != -1:
                    Hydrographs.loc[:, SubsID.loc[i, 0]] = pd.read_csv(
                        HydrologicalInputsPath
                        + "/"
                        + str(int(USnode.loc[SubsID.loc[i, 0] - 1, "US"]))
                        + ".txt"
                    ).values[: len(Hydrographs)]
            except:
                OtherSubLoc = np.where(ReplacementSub["missing"] == SubsID.loc[i, 0])[
                    0
                ][0]
                if (
                    int(
                        USnode.loc[
                            ReplacementSub.loc[OtherSubLoc, "replacement"] - 1, "US"
                        ]
                    )
                    != -1
                ):
                    Hydrographs.loc[:, SubsID.loc[i, 0]] = pd.read_csv(
                        HydrologicalInputsPath
                        + "/"
                        + str(
                            int(
                                USnode.loc[
                                    ReplacementSub.loc[OtherSubLoc, "replacement"] - 1,
                                    "US",
                                ]
                            )
                        )
                        + ".txt"
                    ).values[: len(Hydrographs)]

        # read sub basin map id
        SubIDMap = gdal.Open(SubIDMapF)
        SubIDMapV = SubIDMap.ReadAsArray()
        # NoValue = SubIDMap.GetRasterBand(1).GetNoDataValue()
        # SubIDMapV[SubIDMapV == NoValue] = 0
        # plt.imshow(SubIDMapV)

        # read the added subs reference text file
        ExtraSubs = pd.read_csv(ExtraSubsF)

        # function to write the numbers in the ASCII file

        # read Max depth map
        check = list()
        Klist = list()

        if Tofile == "end" or Tofile > len(MaxDepthList):
            Tofile = len(MaxDepthList)

        # Fromfile = 48
        # Tofile = Fromfile +1

        for k in range(Fromfile, Tofile):

            try:
                # open the zip file
                Compressedfile = zipfile.ZipFile(MapsPath + "/" + MaxDepthList[k])
            except:
                print("Error Opening the compressed file")
                check.append(MaxDepthList[k][len(prefix) : -4])
                Klist.append(k)
                continue

            # get the file name
            fname = Compressedfile.infolist()[0]
            # get the time step from the file name
            timestep = int(fname.filename[len(prefix) : -4])
            print("File= " + str(timestep))

            ASCIIF = Compressedfile.open(fname)
            f = ASCIIF.readlines()
            SpatialRef = f[:6]
            ASCIIRaw = f[6:]
            # ASCIIF = Compressedfile.open(fname)
            # ASCIIRaw = ASCIIF.readlines()[6:]
            rows = len(ASCIIRaw)
            cols = len(ASCIIRaw[0].split())
            MaxDepth = np.ones((rows, cols), dtype=np.float32)
            # read the ascii file
            for i in range(rows):
                x = ASCIIRaw[i].split()
                MaxDepth[i, :] = list(map(float, x))

            # check on the values of the water depth
            #    if np.shape(MaxDepth[np.isnan(MaxDepth)])[0] > 0:
            #        check.append(timestep)
            #        print("Error Check Max Depth values")
            #        continue

            # plotting to check values
            #    fromrow = np.where(MaxDepth == MaxDepth.max())[0][0]
            #    fromcol = np.where(MaxDepth == MaxDepth.max())[1][0]
            #    plt.imshow(MaxDepth[fromrow-20:fromrow+20,fromcol-20:fromcol+20])
            #    plt.imshow(MaxDepth)
            #    plt.colorbar()

            # get the Peak of the hydrograph for the whole event
            # (14 days before the end of the event)
            MaxValuedf = Hydrographs.loc[timestep - 14 : timestep, :]
            MaxValues = MaxValuedf.max().values.tolist()
            T = list()

            # Calculate the the Return period for the max Q at this time step for each
            for i in range(len(MaxValues)):
                # if the sub basin is a lateral and not routed in RIM it will not have a
                # hydrograph
                if np.isnan(MaxValues[i]):
                    T.append(np.nan)
                if not np.isnan(MaxValues[i]):
                    # np.where(USnode['SubID'] == SubsID.loc[i,0])
                    try:
                        DSnode = USnode.loc[SubsID.loc[i, 0] - 1, "US"]
                        loc = np.where(DistributionPr["id"] == DSnode)[0][0]
                    except IndexError:
                        OtherSubLoc = np.where(
                            ReplacementSub["missing"] == SubsID.loc[i, 0]
                        )[0][0]
                        DSnode = USnode.loc[
                            ReplacementSub.loc[OtherSubLoc, "replacement"] - 1, "US"
                        ]
                        loc = np.where(DistributionPr["id"] == DSnode)[0][0]

                    # to get the Non Exceedance probability for a specific Value
                    F = gumbel_r.cdf(
                        MaxValues[i],
                        loc=DistributionPr.loc[loc, "loc"],
                        scale=DistributionPr.loc[loc, "scale"],
                    )
                    # then calculate the the T (return period) T = 1/(1-F)
                    T.append(round(1 / (1 - F), 2))

            try:
                RetunPeriodMap = np.ones((rows, cols), dtype=np.float32) * 0
                for i in range(rows):
                    for j in range(cols):
                        # print("i = " + str(i) + ", j= " + str(j))
                        if not np.isnan(MaxDepth[i, j]):
                            if MaxDepth[i, j] > 0:
                                # print("i = " + str(i) + ", j= " + str(j))
                                # if the sub basin is in the Sub ID list
                                if SubIDMapV[i, j] in SubsID[0].tolist():
                                    # print("Sub = " + str(SubIDMapV[i,j]))
                                    # go get the return period directly
                                    RetunPeriodMap[i, j] = T[
                                        np.where(SubsID[0] == SubIDMapV[i, j])[0][0]
                                    ]
                                else:
                                    # print("Extra  Sub = " + str(SubIDMapV[i,j]))
                                    # the sub ID is one of the added subs not routed by RIM
                                    # so it existed in the ExtraSubs list with a reference to
                                    # a SubID routed by RIM
                                    RIMSub = ExtraSubs.loc[
                                        np.where(
                                            ExtraSubs["addSub"] == SubIDMapV[i, j]
                                        )[0][0],
                                        "RIMSub",
                                    ]
                                    RetunPeriodMap[i, j] = T[
                                        np.where(SubsID[0] == RIMSub)[0][0]
                                    ]
            except:
                print("Error")
                check.append(timestep)
                Klist.append(k)
                continue

            # save the return period ASCII file
            fname = "ReturnPeriod" + str(timestep) + ".asc"

            with open(SaveTo + "/" + fname, "w") as File:
                # write the first lines
                for i in range(len(SpatialRef)):
                    File.write(str(SpatialRef[i].decode()[:-2]) + "\n")

                for i in range(np.shape(RetunPeriodMap)[0]):
                    File.writelines(list(map(self.StringSpace, RetunPeriodMap[i, :])))
                    File.write("\n")

            # zip the file
            with zipfile.ZipFile(
                SaveTo + "/" + fname[:-4] + ".zip", "w", zipfile.ZIP_DEFLATED
            ) as newzip:
                newzip.write(SaveTo + "/" + fname, arcname=fname)
            # delete the file
            os.remove(SaveTo + "/" + fname)

        check = list(zip(check, Klist))
        if len(check) > 0:
            np.savetxt(wpath + "CheckWaterDepth.txt", check, fmt="%6d")

    def CreateTraceALL(
        self,
        ConfigFilePath,
        RIMSubsFilePath,
        TraceFile,
        USonly=1,
        HydrologicalInputsFile="",
    ):
        """CreateTraceALL.

        Parameters
        ----------
        ConfigFilePath : [String]
            SWIM configuration file.
        RIMSubsFilePath : [String]
            path to text file with all the ID of SWIM sub-basins.
        TraceFile : TYPE
            DESCRIPTION.
        USonly : TYPE, optional
            DESCRIPTION. The default is 1.
        HydrologicalInputsFile : TYPE, optional
            DESCRIPTION. The default is ''.

        Returns
        -------
        None.
        """
        # reading the file
        Config = pd.read_csv(ConfigFilePath, header=None)
        # process the Configuration file
        # get the Route rows from the file
        Route = pd.DataFrame(columns=["No", "DSnode", "SubID", "USnode", "No2"])

        j = 0
        for i in range(len(Config)):
            if Config[0][i].split()[0] == "route":
                Route.loc[j, :] = list(map(int, Config[0][i].split()[1:]))
                j = j + 1

        # get RIM Sub-basins
        Subs = pd.read_csv(RIMSubsFilePath, header=None)
        Subs = Subs.rename(columns={0: "SubID"})

        Subs["US"] = None
        Subs["DS"] = None

        for i in range(len(Subs)):
            try:
                # if the sub-basin is in the route array so it is routed by SWIM
                loc = np.where(Route["SubID"] == Subs.loc[i, "SubID"])[0][0]
                Subs.loc[i, "US"] = int(Route.loc[loc, "USnode"])
                Subs.loc[i, "DS"] = int(Route.loc[loc, "DSnode"])
            except IndexError:
                # if the sub-basin is not in the route array so it is not routed by SWIM
                # but still can be routed using RIM

                # Subs.loc[i,'US'] = None
                # Subs.loc[i,'DS'] = None
                Subs.loc[i, "US"] = -1
                Subs.loc[i, "DS"] = -1

        # Save the file with the same format required for the hg R code

        Subs.to_csv(TraceFile, index=None, header=True)
        #    ToSave = Subs.loc[:,['SubID','US']]
        #    ToSave['Extra column 1'] = -1
        #    ToSave['Extra column 2'] = -1
        #    ToSave.to_csv(SavePath + TraceFile,header = None, index = None)

    def ListAttributes(self):
        """ListAttributes.

        Print Attributes List
        """

        print("\n")
        print(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")



"""DistParameters The module contains functions that is responible for distributing parameters spatially (totally distributed, totally distriby-uted with some parameters lumped, all parameters are lumped, hydrologic response units) and also save generated parameters into rasters.

@author: Mostafa
"""
import datetime as dt
import math

# import numbers
import os

import numpy as np
from osgeo import gdal
from pyramids.catchment import Catchment as GC
from pyramids.raster import Raster


class DistParameters:
    """Distripute parameter class is used to distribute the values of the parameter vector in the calibration process into the 3D array, considering if some of the parameters are lumped parameters, if you want to distribute the parameters in HRUs.

    the method included are
    1- par3d
    2- par3dLumped
    3- par2dLumpedK1_lake
    4- HRU
    5- HRU_HAND
    6- ParametersNumber
    7- saveParameters
    """

    def __init__(
        self,
        raster,
        no_parameters,
        no_lumped_par=0,
        lumped_par_pos=[],
        Lake=False,
        Snow=False,
        HRUs=False,
        Function=1,
        Kub=1,
        Klb=50,
        Maskingum=False,
    ):
        """DistParameters. To initiate the DistParameters class you have to provide the Flow Acc raster.

        Parameters
        ----------
        raster : [gdal.dataset]
            raster to get the spatial information of the catchment
            (DEM, flow accumulation or flow direction raster)
        no_parameters : [integer]
            Number of parameters in the HBV model.
        no_lumped_par : [integer], optional
            Number lumped parameters. The default is 0.
        lumped_par_pos : [integer], optional
            the order of the lumped parameters (order starts wti zero). The default is [].
        Lake : [integer], optional
            0 if there is no lake and 1 if there is a lake. The default is 0.
        Snow : [integer]
            0 if you dont want to run the snow related processes and 1 if there is snow.
            in case of 1 (simulate snow processes) parameters related to snow simulation
            has to be provided. The default is 0.
        HRUs : [integer], optional
            if the will consider using HRUs. The default is 0.
        Function : [integer], optional
            function you use to distribute parameters. The default is 1.
        Kub : [numeric], optional
            upper bound for the k-muskingum parameter. The default is 1.
        Klb : [numeric], optional
            lower bound for the k-muskingum parameter. The default is 0.5.
        Maskingum : [bool], optional
            if the routing function is muskingum. The default is False.

        Returns
        -------
        None.
        """
        assert isinstance(raster, gdal.Dataset), (
            "raster should be read using gdal (gdal dataset please read it "
            "using gdal library) "
        )
        assert isinstance(no_parameters, int), " no_parameters should be integer number"
        assert isinstance(
            no_lumped_par, int
        ), "no of lumped parameters should be integer"

        if no_lumped_par >= 1:
            if isinstance(lumped_par_pos, list):
                assert no_lumped_par == len(lumped_par_pos), (
                    f"you have to entered {no_lumped_par} no of lumped parameters but only {len(lumped_par_pos)} "
                    f"position "
                )
            else:  # if not int or list
                raise ValueError(
                    "you have one or more lumped parameters so the position has to be entered as a list"
                )

        self.Lake = Lake
        self.Snow = Snow
        self.no_lumped_par = no_lumped_par
        self.lumped_par_pos = lumped_par_pos
        self.HRUs = HRUs
        self.Kub = Kub
        self.Klb = Klb
        self.Maskingum = Maskingum
        # read the raster
        self.raster = raster
        self.raster_A = raster.ReadAsArray().astype(float)
        # get the shape of the raster
        self.rows = raster.RasterYSize
        self.cols = raster.RasterXSize
        # get the no_value of in the raster
        self.noval = raster.GetRasterBand(1).GetNoDataValue()

        for i in range(self.rows):
            for j in range(self.cols):
                if math.isclose(self.raster_A[i, j], self.noval, rel_tol=0.001):
                    self.raster_A[i, j] = np.nan

        # count the number of non-empty cells
        if self.HRUs:
            self.values = list(
                set(
                    [
                        int(self.raster_A[i, j])
                        for i in range(self.rows)
                        for j in range(self.cols)
                        if not np.isnan(self.raster_A[i, j])
                    ]
                )
            )
            self.no_elem = len(self.values)
        else:
            self.no_elem = np.size(self.raster_A[:, :]) - np.count_nonzero(
                (self.raster_A[np.isnan(self.raster_A)])
            )

        self.no_parameters = no_parameters

        # store the indeces of the non-empty cells
        self.celli = []
        self.cellj = []
        for i in range(self.rows):
            for j in range(self.cols):
                if not np.isnan(self.raster_A[i, j]):
                    self.celli.append(i)
                    self.cellj.append(j)

        # create an empty 3D array [[raster dimension], no_parameters]
        self.Par3d = np.zeros([self.rows, self.cols, self.no_parameters]) * np.nan

        if no_lumped_par >= 1:
            # parameters in array
            # remove a place for the lumped parameter (k1) lower zone coefficient
            self.no_parameters = self.no_parameters - no_lumped_par

        # all parameters lumped and distributed
        self.totnumberpar = self.no_parameters * self.no_elem + no_lumped_par
        # parameters in array
        # create a 2d array [no_parameters, no_cells]
        self.Par2d = np.zeros(shape=(self.no_parameters, self.no_elem), dtype=np.float)

        if Function == 1:
            self.Function = self.par3dLumped
        elif Function == 2:
            self.Function = self.par3d
        elif Function == 3:
            self.Function = self.par2d_lumpedK1_lake
        elif Function == 4:
            self.Function = self.HRU
        # to overwrite any choice user choose if the is HRUs
        if self.HRUs == 1:
            self.Function = self.HRU

        self.ParametersNumber()

        pass

    def par3d(self, par_g):  # , kub=1,klb=0.5,Maskingum=True
        """par3d method takes a list of parameters [saved as one column or generated as 1D list from optimization algorithm] and distribute them horizontally on number of cells given by a raster.

        Parameters
        ----------
        par_g : [list]
            list of parameters
        raster : [gdal.dataset]
            raster to get the spatial information of the catchment
            (DEM, flow accumulation or flow direction raster)
        no_parameters : [integer]
            no of parameters of the cell according to the rainfall runoff model
        no_lumped_par : [integer]
            nomber of lumped parameters, you have to enter the value of
            the lumped parameter at the end of the list, default is 0 (no lumped parameters)
        lumped_par_pos : [List]
            list of order or position of the lumped parameter among all
            the parameters of the lumped model (order starts from 0 to the length
            of the model parameters), default is [] (empty), the following order
            of parameters is used for the lumped HBV model used
            [ltt, utt, rfcf, sfcf, ttm, cfmax, cwh, cfr, fc, beta, e_corr, etf, lp,
            c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
        kub : [float]
            upper bound of K value (traveling time in muskingum routing method)
            default is 1 hour
        klb : [float]
            Lower bound of K value (traveling time in muskingum routing method)
            default is 0.5 hour (30 min)
        Maskingum : [bool], optional
            if the routing function is muskingum. The default is False.

        Returns
        -------
        par_3d : [3d array]
            3D array of the parameters distributed horizontally on the cells

        Examples
        --------
        EX1:totally distributed parameters
            raster=gdal.Open("dem.tif")
            [fc, beta, etf, lp, c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
            no_lumped_par=0
            lumped_par_pos=[]
            par_g=np.random.random(no_elem*(no_parameters-no_lumped_par))

            tot_dist_par=par3d(par_g,raster,no_parameters,no_lumped_par,lumped_par_pos,kub=1,klb=0.5)

        EX2: One Lumped Parameter [K1]
            raster=gdal.Open("dem.tif")
            given values of parameters are of this order
            [fc, beta, etf, lp, c_flux, k, alpha, perc, pcorr, Kmuskingum, Xmuskingum,k1]
            K1 is lumped so its value is inserted at the end and its order should
            be after K

            no_lumped_par=1
            lumped_par_pos=[6]
            par_g=np.random.random(no_elem* (no_parameters-no_lumped_par))
            # insert the value of k1 at the end
            par_g=np.append(par_g,0.005)

            dist_par=par3d(par_g,raster,no_parameters,no_lumped_par,lumped_par_pos,kub=1,klb=0.5)

        EX3:Two Lumped Parameter [K1, Perc]
            raster=gdal.Open("dem.tif")
            no_lumped_par=2
            lumped_par_pos=[6,8]
            par_g=np.random.random(no_elem* (no_parameters-no_lumped_par))
            par_g=np.append(par_g,0.005)
            par_g=np.append(par_g,0.006)

            dist_par=par3d(par_g,raster,no_parameters,no_lumped_par,lumped_par_pos,kub=1,klb=0.5)
        """
        # input data validation
        # data type
        # assert type(par_g)==np.ndarray or type(par_g)==list, "par_g should be of type 1d array or list"
        # assert isinstance(kub,numbers.Number) , " kub should be a number"
        # assert isinstance(klb,numbers.Number) , " klb should be a number"

        # input values
        if self.no_lumped_par > 0:
            par_no = (self.no_elem * self.no_parameters) + self.no_lumped_par
            assert len(par_g) == par_no, (
                f"As there is {self.no_lumped_par} lumped parameters, length of input parameters should be {self.no_elem}"
                + f"*({self.no_parameters + self.no_lumped_par} - {self.no_lumped_par}) + {self.no_lumped_par} = "
                + f"{self.no_elem * (self.no_parameters - self.no_lumped_par) + self.no_lumped_par} not {len(par_g)}"
                + " probably you have to add the value of the lumped parameter at the end of the list"
            )
        else:
            # if there is no lumped parameters
            par_no = self.no_elem * self.no_parameters
            assert len(par_g) == par_no, (
                f"As there is no lumped parameters length of input parameters should be {self.no_elem} * "
                + f"{self.no_parameters} = {self.no_elem * self.no_parameters}"
            )

        # parameters in array
        # create a 2d array [no_parameters, no_cells]
        self.Par2d = np.ones((self.no_parameters, self.no_elem))
        # take the parameters from the generated parameters or the 1D list and
        # assign them to each cell
        for i in range(self.no_elem):
            self.Par2d[:, i] = par_g[
                i * self.no_parameters : (i * self.no_parameters) + self.no_parameters
            ]

        ### lumped parameters
        if self.no_lumped_par > 0:
            for i in range(self.no_lumped_par):
                # create a list with the value of the lumped parameter(k1)
                # (stored at the end of the list of the parameters)
                pk1 = (
                    np.ones((1, self.no_elem))
                    * par_g[(self.no_parameters * np.shape(self.Par2d)[1]) + i]
                )
                # put the list of parameter k1 at the 6 row
                self.Par2d = np.vstack(
                    [
                        self.Par2d[: self.lumped_par_pos[i], :],
                        pk1,
                        self.Par2d[self.lumped_par_pos[i] :, :],
                    ]
                )

        # assign the parameters from the array (no_parameters, no_cells) to
        # the spatially corrected location in par2d
        for i in range(self.no_elem):
            self.Par3d[self.celli[i], self.cellj[i], :] = self.Par2d[:, i]

        # calculate the value of k(travelling time in muskingum based on value of
        # x and the position and upper, lower bound of k value

        # if Maskingum:
        #     for i in range(self.no_elem):
        #         self.Par3d[self.celli[i],self.cellj[i],-2]= DistParameters.calculateK(self.Par3d[self.celli[i],self.cellj[i],-1],self.Par3d[self.celli[i],self.cellj[i],-2],kub,klb)

    def par3dLumped(self, par_g):  # , kub=1, klb=0.5, Maskingum = True
        """par3dLumped method takes a list of parameters [saved as one column or generated as 1D list from optimization algorithm] and distribute them horizontally on number of cells given by a raster.

        Parameters
        ----------
        par_g : [list]
            list of parameters
        kub : [float]
            upper bound of K value (traveling time in muskingum routing method)
            default is 1 hour
        klb : [float]
            Lower bound of K value (traveling time in muskingum routing method)
            default is 0.5 hour (30 min)
        Maskingum : [bool], optional
            if the routing function is muskingum. The default is False.

        Returns
        ----------
        par_3d: [3d array]
            3D array of the parameters distributed horizontally on the cells

        Example:
        ----------
        EX1:Lumped parameters
            raster=gdal.Open("dem.tif")
            [fc, beta, etf, lp, c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]

            raster=gdal.Open(path+"soil_classes.tif")
            no_parameters=12
            par_g=np.random.random(no_parameters) #no_elem*(no_parameters-no_lumped_par)

            tot_dist_par=DP.par3dLumped(par_g,raster,no_parameters,lumped_par_pos,kub=1,klb=0.5)
        """
        # input data validation
        # data type
        assert (
            type(par_g) == np.ndarray or type(par_g) == list
        ), "par_g should be of type 1d array or list"
        # assert isinstance(kub,numbers.Number) , " kub should be a number"
        # assert isinstance(klb,numbers.Number) , " klb should be a number"

        # take the parameters from the generated parameters or the 1D list and
        # assign them to each cell
        for i in range(self.no_elem):
            self.Par2d[:, i] = par_g

        # assign the parameters from the array (no_parameters, no_cells) to
        # the spatially corrected location in par2d
        for i in range(self.no_elem):
            self.Par3d[self.celli[i], self.cellj[i], :] = self.Par2d[:, i]

        # calculate the value of k(travelling time in muskingum based on value of
        # x and the position and upper, lower bound of k value
        # if Maskingum == True:
        #     for i in range(self.no_elem):
        #         self.Par3d[self.celli[i],self.cellj[i],-2] = DistParameters.calculateK(self.Par3d[self.celli[i],self.cellj[i],-1],
        #                                                                                self.Par3d[self.celli[i],self.cellj[i],-2],
        #                                                                                kub,klb)

    @staticmethod
    def calculateK(x, position, UB, LB):
        """calculateK. calculateK method takes value of x parameter and generate 100 random value of k parameters between upper & lower constraint then the output will be the value coresponding to the giving position.

        Parameters
        ----------
        x : [numeric]
            weighting coefficient to determine the linearity of the water surface
            (one of the parameters of muskingum routing method)
        position : [integer]
            random position between upper and lower bounds of the k parameter
        UB : [numeric]
            upper bound for k parameter
        LB : [numeric]
            Lower bound for k parameter
        """
        # k has to be smaller than this constraint
        constraint1 = 0.5 * 1 / (1 - x)
        # k has to be greater than this constraint
        constraint2 = 0.5 * 1 / x
        # if constraint is higher than UB take UB
        if constraint2 >= UB:
            constraint2 = UB
        # if constraint is lower than LB take UB
        if constraint1 <= LB:
            constraint1 = LB

        generatedK = np.linspace(constraint1, constraint2, 50)
        k = generatedK[int(round(position, 0))]
        return k

    def par2d_lumpedK1_lake(self, par_g, no_parameters_lake):  # ,kub,klb
        """par2d_lumpedK1_lake method takes a list of parameters and distribute them horizontally on number of cells given by a raster.

        Parameters
        ----------
        par_g : [list]
            list of parameters
        raster : [gdal.dataset]
            raster to get the spatial information of the catchment
            (DEM, flow accumulation or flow direction raster)
        no_parameters :[integer]
            no of parameters of the cell
        no_parameters_lake : [integer]
            no of lake parameters
        kub : [float]
            upper bound of K value (traveling time in muskingum routing method)
            default is 1 hour
        klb : [float]
            Lower bound of K value (traveling time in muskingum routing method)
            default is 0.5 hour (30 min)

        Returns
        -------
        Par3d: [3d array]
            3D array of the parameters distributed horizontally on the cells
        lake_par: [list]
            list of the lake parameters.

        Example:
        ----------
        a list of 155 value,all parameters are distributed except lower zone coefficient
        (is written at the end of the list) each cell(14 cells) has 11 parameter plus lower zone
        (12 parameters) function will take each 11 parameter and assing them to a specific cell
        then assign the last value (lower zone parameter) to all cells
        14*11=154 + 1 = 155
        """
        # parameters in array
        # remove a place for the lumped parameter (k1) lower zone coefficient
        no_parameters = self.no_parameters - 1

        # create a 2d array [no_parameters, no_cells]
        self.Par2d = np.ones((no_parameters, self.no_elem))

        # take the parameters from the generated parameters or the 1D list and
        # assign them to each cell
        for i in range(self.no_elem):
            self.Par2d[:, i] = par_g[
                i * no_parameters : (i * no_parameters) + no_parameters
            ]

        # create a list with the value of the lumped parameter(k1)
        # (stored at the end of the list of the parameters)
        pk1 = (
            np.ones((1, self.no_elem))
            * par_g[(np.shape(self.Par2d)[0] * np.shape(self.Par2d)[1])]
        )

        # put the list of parameter k1 at the 6 row
        self.Par2d = np.vstack([self.Par2d[:6, :], pk1, self.Par2d[6:, :]])

        # assign the parameters from the array (no_parameters, no_cells) to
        # the spatially corrected location in par2d
        for i in range(self.no_elem):
            self.Par3d[self.celli[i], self.cellj[i], :] = self.Par2d[:, i]

        # calculate the value of k(travelling time in muskingum based on value of
        # x and the position and upper, lower bound of k value
        # for i in range(self.no_elem):
        #     self.Par3d[self.celli[i],self.cellj[i],-2]= DistParameters.calculateK(self.Par3d[self.celli[i],self.cellj[i],-1],self.Par3d[self.celli[i],self.cellj[i],-2],kub,klb)

        # lake parameters
        self.lake_par = par_g[len(par_g) - no_parameters_lake :]
        # self.lake_par[-2] = DistParameters.calculateK(self.lake_par[-1],self.lake_par[-2],kub,klb)

        # return self.Par3d, lake_par

    def HRU(self, par_g):  # ,kub=1,klb=0.5
        """HRU method takes a list of parameters [saved as one column or generated as 1D list from optimization algorithm] and distribute them horizontally on number of cells given by a raster the input raster should be classified raster (by numbers) into class to be used to define the HRUs.

        Parameters
        ----------
        par_g:
            [list] list of parameters
        raster:
            [gdal.dataset] classification raster to get the spatial information
            of the catchment and the to define each cell belongs to which HRU
        no_parameters
            [int] no of parameters of the cell according to the rainfall runoff model
        no_lumped_par:
            [int] nomber of lumped parameters, you have to enter the value of
            the lumped parameter at the end of the list, default is 0 (no lumped parameters)
        lumped_par_pos:
            [List] list of order or position of the lumped parameter among all
            the parameters of the lumped model (order starts from 0 to the length
            of the model parameters), default is [] (empty), the following order
            of parameters is used for the lumped HBV model used
            [ltt, utt, rfcf, sfcf, ttm, cfmax, cwh, cfr, fc, beta, e_corr, etf, lp,
            c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
        kub:
            [float] upper bound of K value (traveling time in muskingum routing method)
            default is 1 hour
        klb:
            [float] Lower bound of K value (traveling time in muskingum routing method)
            default is 0.5 hour (30 min)

        Returns
        ----------
        par_3d:
            3D array of the parameters distributed horizontally on the cells

        Examples
        --------
        EX1:HRU without lumped parameters
            [fc, beta, etf, lp, c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
            raster = gdal.Open("soil_types.tif")
            no_lumped_par=0
            lumped_par_pos=[]
            par_g=np.random.random(no_elem*(no_parameters-no_lumped_par))

            par_hru=HRU(par_g,raster,no_parameters,no_lumped_par,lumped_par_pos,kub=1,klb=0.5)


        EX2: HRU with one lumped parameters
            given values of parameters are of this order
            [fc, beta, etf, lp, c_flux, k, alpha, perc, pcorr, Kmuskingum, Xmuskingum,k1]
            K1 is lumped so its value is inserted at the end and its order should
            be after K

            raster = gdal.Open("soil_types.tif")
            no_lumped_par=1
            lumped_par_pos=[6]
            par_g=np.random.random(no_elem* (no_parameters-no_lumped_par))
            # insert the value of k1 at the end
            par_g=np.append(par_g,0.005)

            par_hru=HRU(par_g,raster,no_parameters,no_lumped_par,lumped_par_pos,kub=1,klb=0.5)
        """
        # input data validation
        # data type
        assert (
            type(par_g) == np.ndarray or type(par_g) == list
        ), "par_g should be of type 1d array or list"
        # assert isinstance(kub,numbers.Number) , " kub should be a number"
        # assert isinstance(klb,numbers.Number) , " klb should be a number"

        # input values
        if self.no_lumped_par > 0:
            par_no = (self.no_elem * self.no_parameters) + self.no_lumped_par
            assert len(par_g) == par_no, (
                f"As there is {self.no_lumped_par} lumped parameters, length of input parameters should be "
                f"{self.no_elem}*({self.no_parameters}-{self.no_lumped_par})+{self.no_lumped_par}="
                + str(
                    self.no_elem * (self.no_parameters - self.no_lumped_par)
                    + self.no_lumped_par
                )
                + f" not {len(par_g)} probably you have to add the value of the lumped parameter at the end of the list"
            )
        else:
            # if there is no lumped parameters
            assert len(par_g) == self.no_elem * self.no_parameters, (
                f"As there is no lumped parameters length of input parameters should be {self.no_elem}*{self.no_parameters}"
                + f"={self.no_elem * self.no_parameters}"
            )

        # take the parameters from the generated parameters or the 1D list and
        # assign them to each cell
        self.Par2d = np.zeros(shape=(self.no_parameters, self.no_elem), dtype=np.float)
        for i in range(self.no_elem):
            self.Par2d[:, i] = par_g[
                i * self.no_parameters : (i * self.no_parameters) + self.no_parameters
            ]

        ### lumped parameters
        if self.no_lumped_par > 0:
            for i in range(self.no_lumped_par):
                # create a list with the value of the lumped parameter(k1)
                # (stored at the end of the list of the parameters)
                pk1 = (
                    np.ones((1, self.no_elem))
                    * par_g[(self.no_parameters * np.shape(self.Par2d)[1]) + i]
                )
                # put the list of parameter k1 at the 6 row
                self.Par2d = np.vstack(
                    [
                        self.Par2d[: self.lumped_par_pos[i], :],
                        pk1,
                        self.Par2d[self.lumped_par_pos[i] :, :],
                    ]
                )

        # calculate the value of k(travelling time in muskingum based on value of
        # x and the position and upper, lower bound of k value
        # for i in range(self.no_elem):
        #     self.Par2d[-2,i] = DistParameters.calculateK(self.Par2d[-1,i],self.Par2d[-2,i],kub,klb)

        # assign the parameters from the array (no_parameters, no_cells) to
        # the spatially corrected location in par2d each soil type will have the same
        # generated parameters
        for i in range(self.no_elem):
            self.Par3d[self.raster_A == self.values[i]] = self.Par2d[:, i]

    @staticmethod
    def HRU_HAND(DEM, FD, FPL, River):
        """HRU_HAND this function calculates inputs for the HAND (height above nearest drainage) method for land use classification.

        Parameters
        ----------
        DEM :
            raster to get the spatial information of the catchment
            (DEM raster)
        FD : [gdal.dataset]
            flow direction  raster to get the spatial information of the catchment
        FPL : [gdal.dataset]
            raster to get the spatial information of the catchment
        River : [gdal.dataset]
            raster to get the spatial information of the catchment


        Returns
        -------
        HAND: [numpy ndarray]
            Height above nearest drainage

        DTND: [numpy ndarray]
            Distance to nearest drainage
        """
        # Use DEM raster information to run all loops
        dem_A = DEM.ReadAsArray()
        no_val = np.float32(DEM.GetRasterBand(1).GetNoDataValue())
        rows = DEM.RasterYSize
        cols = DEM.RasterXSize

        # get the indices of the flow direction path
        fd_index = GC.flowDirectionIndex(FD)

        # read the river location raster
        river_A = River.ReadAsArray()

        # read the flow path length raster
        fpl_A = FPL.ReadAsArray()

        # trace the flow direction to the nearest river reach and store the location
        # of that nearst reach
        nearest_network = np.ones((rows, cols, 2)) * np.nan
        try:
            for i in range(rows):
                for j in range(cols):
                    if dem_A[i, j] != no_val:
                        f = river_A[i, j]
                        old_row = i
                        old_cols = j

                        while f != 1:
                            # did not reached to the river yet then go to the next down stream cell
                            # get the down stream cell (furure position)
                            new_row = int(fd_index[old_row, old_cols, 0])
                            new_cols = int(fd_index[old_row, old_cols, 1])
                            # print(str(new_row)+","+str(new_cols))
                            # go to the downstream cell
                            f = river_A[new_row, new_cols]
                            # down stream cell becomes the current position (old position)
                            old_row = new_row
                            old_cols = new_cols
                            # at this moment old and new stored position are the same (current position)
                        # store the position in the array
                        nearest_network[i, j, 0] = new_row
                        nearest_network[i, j, 1] = new_cols

        except:
            assert (
                False
            ), "please check the boundaries of your catchment after cropping the catchment using the a polygon it creates anomalies athe boundary "

        # calculate the elevation difference between the cell and the nearest drainage cell
        # or height avove nearst drainage
        HAND = np.ones((rows, cols)) * np.nan

        for i in range(rows):
            for j in range(cols):
                if dem_A[i, j] != no_val:
                    HAND[i, j] = (
                        dem_A[i, j]
                        - dem_A[
                            int(nearest_network[i, j, 0]), int(nearest_network[i, j, 1])
                        ]
                    )

        # calculate the distance to the nearest drainage cell using flow path length
        # or distance to nearest drainage
        DTND = np.ones((rows, cols)) * np.nan

        for i in range(rows):
            for j in range(cols):
                if dem_A[i, j] != no_val:
                    DTND[i, j] = (
                        fpl_A[i, j]
                        - fpl_A[
                            int(nearest_network[i, j, 0]), int(nearest_network[i, j, 1])
                        ]
                    )

        return HAND, DTND

    def ParametersNumber(self):
        """ParametersNumber. ParametersNO method calculates the nomber of parameters that the optimization algorithm is going top search for, use it only in case of totally distributed catchment parameters (in case of lumped parameters no of parameters are the same as the no of parameters of the conceptual model)

        Parameters
        ----------
        raster : [gdal.dataset]
            raster to get the spatial information of the catchment
            (DEM, flow accumulation or flow direction raster)
        no_parameters : [integer]
            no of parameters of the cell according to the rainfall runoff model
        no_lumped_par : [integer]
            nomber of lumped parameters, you have to enter the value of
            the lumped parameter at the end of the list, default is 0 (no lumped parameters)
        HRUs : [0 or 1]
            0 to define that no hydrologic response units (HRUs), 1 to define that
            HRUs are used
        """
        if not self.HRUs:
            if self.no_lumped_par > 0:
                # self.ParametersNO = (self.no_elem *( self.no_parameters - self.no_lumped_par)) + self.no_lumped_par
                self.ParametersNO = (
                    self.no_elem * self.no_parameters
                ) + self.no_lumped_par
            else:
                # if there is no lumped parameters
                self.ParametersNO = self.no_elem * self.no_parameters
        else:
            if self.no_lumped_par > 0:
                # self.ParametersNO = (self.no_elem * (self.no_parameters - self.no_lumped_par)) + self.no_lumped_par
                self.ParametersNO = (
                    self.no_elem * self.no_parameters
                ) + self.no_lumped_par
            else:
                # if there is no lumped parameters
                self.ParametersNO = self.no_elem * self.no_parameters

    def saveParameters(self, Path):
        """saveParameters. saveParameters method takes generated parameters by the calibration algorithm, distributed them with a given function and save them as a rasters.

         Parameters
         ----------
         DistParFn:
             [function] function to distribute the parameters (all functions are
             in Hapi.DistParameters )
         raster:
             [gdal.dataset] raster to get the spatial information
         Par
             [list or numpy ndarray] parameters as 1D array or list
         no_parameters:
             [int] number of the parameters in the conceptual model
         snow:
             [integer] number to define whether to take parameters of
             the conceptual model with snow subroutine or without
         kub:
             [numeric] upper bound for k parameter in muskingum function
         klb:
             [numeric] lower bound for k parameter in muskingum function
         Path:
              [string] path to the folder you want to save the parameters in
              default value is None (parameters are going to be saved in the
              current directory)

         Returns
         -------
              Rasters for parameters of the distributed model

        examples:
        ----------
         >>> DemPath = path+"GIS/4000/dem4000.tif"
         >>> raster=gdal.Open(DemPath)
         >>> ParPath = "par15_7_2018.txt"
         >>> par=np.loadtxt(ParPath)
         >>> klb=0.5
             kub=1
             no_parameters=12
             DistParFn=DP.par3dLumped
             Path="parameters/"
             snow=0

             saveParameters(DistParFn, raster, par, no_parameters,snow ,kub, klb,Path)
        """
        assert type(Path) == str, "path should be of type string"
        assert os.path.exists(Path), Path + " you have provided does not exist"

        # save
        if self.Snow == 0:  # now snow subroutine
            pnme = [
                "01_rfcf",
                "02_FC",
                "03_BETA",
                "04_ETF",
                "05_LP",
                "06_K0",
                "07_K1",
                "08_K2",
                "09_UZL",
                "10_PERC",
                "11_Kmuskingum",
                "12_Xmuskingum",
            ]
        else:  # there is snow subtoutine
            pnme = [
                "01_ltt",
                "02_rfcf",
                "03_sfcf",
                "04_cfmax",
                "05_cwh",
                "06_cfr",
                "07_fc",
                "08_beta",
                "09_etf",
                "10_lp",
                "11_k0",
                "12_k1",
                "13_k2",
                "14_uzl",
                "18_perc",
            ]

        if Path is not None:
            pnme = [
                Path + i + "_" + str(dt.datetime.now())[0:10] + ".tif" for i in pnme
            ]

        for i in range(np.shape(self.Par3d)[2]):
            Raster.rasterLike(self.raster, self.Par3d[:, :, i], pnme[i])

    def ListAttributes(self):
        """Print Attributes List."""

        print("\n")
        print(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")



"""Calibration.

calibration contains functions to to connect the parameter spatial distribution
function with the with both component of the spatial representation of the hydrological
process (conceptual model & spatial routing) to calculate the performance of predicted
runoff at known locations based on given performance function

@author: Mostafa
"""
import datetime as dt
from typing import Any, Optional

import numpy as np
import pandas as pd
from Oasis.harmonysearch import HSapi
from Oasis.optimization import Optimization

from Hapi.catchment import Catchment
from Hapi.rrm.wrapper import Wrapper


class Calibration(Catchment):
    """Calibration.

    Calibration class contains to connect the parameter spatial distribution
    function with the with both component of the spatial representation of the
    hydrological process (conceptual model & spatial routing) to calculate the
    performance of predicted runoff at known locations based on given
    performance function

    The calibration class is sub-class from the Catchment super class so you
    need to create the Catchment object first to be able to run the calibration
    """

    def __init__(
        self,
        name: Any,
        start: str,
        end: str,
        fmt: str = "%Y-%m-%d",
        SpatialResolution: Optional[str] = "Lumped",
        TemporalResolution: Optional[str] = "Daily",
        RouteRiver: Optional[str] = "Muskingum",
    ):
        """Calibration.

        to instantiate the Calibration object you need to provide the following
        arguments

        Parameters
        ----------
        name : [str]
            Name of the Catchment.
        start : [str]
            starting date.
        end : [str]
            end date.
        fmt : [str], optional
            format of the given date. The default is "%Y-%m-%d".
        SpatialResolution : [str], optional
            Lumped or 'Distributed' . The default is 'Lumped'.
        TemporalResolution : [str], optional
            "Hourly" or "Daily". The default is "Daily".

        Returns
        -------
        None.
        """
        self.name = name
        self.start = dt.datetime.strptime(start, fmt)
        self.end = dt.datetime.strptime(end, fmt)
        self.SpatialResolution = SpatialResolution
        self.TemporalResolution = TemporalResolution

        conversionfactor = (1000 * 24 * 60 * 60) / (1000**2)

        if TemporalResolution.lower() == "daily":
            self.dt = 1  # 24
            self.conversionfactor = conversionfactor * 1
            self.Index = pd.date_range(self.start, self.end, freq="D")
        elif TemporalResolution.lower() == "hourly":
            self.dt = 1  # 24
            self.conversionfactor = conversionfactor * 1 / 24
            self.Index = pd.date_range(self.start, self.end, freq="H")
        else:
            # TODO calculate the teporal resolution factor
            # q mm , area sq km  (1000**2)/1000/f/60/60 = 1/(3.6*f)
            # if daily tfac=24 if hourly tfac=1 if 15 min tfac=0.25
            self.Tfactor = 24

        self.RouteRiver = RouteRiver
        self.RouteRiver = RouteRiver
        self.Parameters = None
        self.data = None
        self.Prec = None
        self.TS = None
        self.Temp = None
        self.ET = None
        self.ll_temp = None
        self.QGauges = None
        self.Snow = None
        self.Maxbas = None
        self.LumpedModel = None
        self.CatArea = None
        self.InitialCond = None
        self.q_init = None
        self.GaugesTable = None
        self.UB = None
        self.LB = None
        self.cols = None
        self.rows = None
        self.NoDataValue = None
        self.FlowAccArr = None
        self.no_elem = None
        self.acc_val = None
        self.Outlet = None
        self.CellSize = None
        self.px_area = None
        self.px_tot_area = None
        self.FlowDirArr = None
        self.FDT = None
        self.FPLArr = None
        (
            self.DEM,
            self.BankfullDepth,
            self.RiverWidth,
            self.RiverRoughness,
            self.FloodPlainRoughness,
        ) = (None, None, None, None, None)
        self.qout, self.Qtot = None, None
        self.quz_routed, self.qlz_translated, self.statevariables = None, None, None
        self.anim = None
        self.quz, self.qlz = None, None
        self.Qsim = None
        self.Metrics = None

        pass

    def readObjectiveFn(self, OF, args):
        """readObjectiveFn.

        readObjectiveFn method takes the objective function and and any arguments
        that are needed to be passed to the objective function.

        Parameters
        ----------
        OF : [function]
            callable function to calculate any kind of metric to be used in the
            calibration.
        args : [positional/keyword arguments]
            any kind of argument you want to pass to your objective function.

        Returns
        -------
        None.
        """
        # check objective_function
        assert callable(OF), "The Objective function should be a function"
        self.OF = OF

        if args is None:
            args = []

        self.OFArgs = args

        print("Objective function is read successfully")

    def extractDischarge(self, Factor=None):
        """extractDischarge.

        extractDischarge method extracts the discharge hydrograph in the
        Q

        Parameters
        ----------
        Factor : [list/None]
            list of factor if you want to multiply the simulated discharge by
            a factor you have to provide a list of the factor (as many factors
            as the number of gauges). The default is False.

        Returns
        -------
        None.
        """
        self.Qsim = np.zeros((self.TS - 1, len(self.GaugesTable)))
        # error = 0
        for i in range(len(self.GaugesTable)):
            Xind = int(self.GaugesTable.loc[self.GaugesTable.index[i], "cell_row"])
            Yind = int(self.GaugesTable.loc[self.GaugesTable.index[i], "cell_col"])
            # gaugeid = self.GaugesTable.loc[self.GaugesTable.index[i],"id"]

            # Quz = self.quz_routed[Xind,Yind,:-1]
            # Qlz = self.qlz_translated[Xind,Yind,:-1]
            # self.Qsim[:,i] = Quz + Qlz

            Qsim = np.reshape(self.Qtot[Xind, Yind, :-1], self.TS - 1)

            if Factor is not None:
                self.Qsim[:, i] = Qsim * Factor[i]
            else:
                self.Qsim[:, i] = Qsim

            # Qobs = Coello.QGauges.loc[:,gaugeid]
            # error = error + OF(Qobs, Qsim)

        # return error

    def runCalibration(self, SpatialVarFun, OptimizationArgs, printError=None):
        """runCalibration.

        this function runs the calibration algorithm for the conceptual distributed
        hydrological model

        Inputs:
        ----------
            1-ConceptualModel:
                [function] conceptual model and it should contain a function called simulate

            2-Basic_inputs:
                1-p2:
                    [List] list of unoptimized parameters
                    p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                    p2[1] = catchment area in km2
                2-init_st:
                    [list] initial values for the state variables [sp,sm,uz,lz,wc] in mm
                3-UB:
                    [Numeric] upper bound of the values of the parameters
                4-LB:
                    [Numeric] Lower bound of the values of the parameters
            3-Q_obs:
                [Numeric] Observed values of discharge

            6-lumpedParNo:
                [int] nomber of lumped parameters, you have to enter the value of
                the lumped parameter at the end of the list, default is 0 (no lumped parameters)
            7-lumpedParPos:
                [List] list of order or position of the lumped parameter among all
                the parameters of the lumped model (order starts from 0 to the length
                of the model parameters), default is [] (empty), the following order
                of parameters is used for the lumped HBV model used
                [ltt, utt, rfcf, sfcf, ttm, cfmax, cwh, cfr, fc, beta, e_corr, etf, lp,
                c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
            8-objective_function:
                [function] objective function to calculate the performance of the model
                and to be used in the calibration
            9-*args:
                other arguments needed on the objective function

        Outputs:
        ----------
            1- st:
                [4D array] state variables
            2- q_out:
                [1D array] calculated Discharge at the outlet of the catchment
            3- q_uz:
                [3D array] Distributed discharge for each cell

        Example:
        ----------
            PrecPath = prec_path="meteodata/4000/calib/prec"
            Evap_Path = evap_path="meteodata/4000/calib/evap"
            TempPath = temp_path="meteodata/4000/calib/temp"
            FlowAccPath = "GIS/4000/acc4000.tif"
            FlowDPath = "GIS/4000/fd4000.tif"
            ParPath = "meteodata/4000/"+"parameters.txt"
            p2=[1, 227.31]
            st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                              FlowAccPath,FlowDPath,ParPath,p2)
        """
        # input dimensions
        # [rows,cols] = self.FlowAcc.ReadAsArray().shape
        [fd_rows, fd_cols] = self.FlowDirArr.shape
        assert (
            fd_rows == self.rows and fd_cols == self.cols
        ), "all input data should have the same number of rows"

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        # basic inputs
        # check if all inputs are included
        # assert all(["p2","init_st","UB","LB","snow "][i] in Basic_inputs.keys() for i in range(4)), "Basic_inputs should contain ['p2','init_st','UB','LB'] "

        ### optimization

        # get arguments
        ApiObjArgs = OptimizationArgs[0]
        pll_type = OptimizationArgs[1]
        ApiSolveArgs = OptimizationArgs[2]
        # check optimization arguement
        assert type(ApiObjArgs) == dict, "store_history should be 0 or 1"
        assert type(ApiSolveArgs) == dict, "history_fname should be of type string "

        print("Calibration starts")

        ### calculate the objective function
        def opt_fun(par):
            try:
                # distribute the parameters
                SpatialVarFun.Function(
                    par
                )  # , kub=SpatialVarFun.Kub, klb=SpatialVarFun.Klb
                self.Parameters = SpatialVarFun.Par3d
                # run the model
                Wrapper.RRMModel(self)
                # calculate performance of the model
                try:
                    error = self.OF(
                        self.QGauges, *[self.GaugesTable]
                    )  # self.qout, self.quz_routed, self.qlz_translated,
                    f = list(range(9, len(par), SpatialVarFun.no_parameters))
                    g = list()
                    for i in range(len(f)):
                        k = par[f[i]]
                        x = par[f[i] + 1]
                        g.append(2 * k * x / self.dt)
                        g.append((2 * k * (1 - x)) / self.dt)

                except TypeError:  # if no of inputs less than what the function needs
                    assert (
                        False
                    ), "the objective function you have entered needs more inputs please enter then in a list as *args"

                # print error
                if printError != 0:
                    print(round(error, 3))
                    print(par)

                fail = 0
            except:
                error = np.nan
                g = []
                fail = 1

            return error, g, fail

        ### define the optimization components
        opt_prob = Optimization("HBV Calibration", opt_fun)
        for i in range(len(self.LB)):
            opt_prob.addVar(
                "x{0}".format(i), type="c", lower=self.LB[i], upper=self.UB[i]
            )

        opt_prob.addObj("f")

        for i in range(SpatialVarFun.no_elem):
            opt_prob.addCon("g" + str(i) + "-1", "i")
            opt_prob.addCon("g" + str(i) + "-2", "i")

        print(opt_prob)

        opt_engine = HSapi(pll_type=pll_type, options=ApiObjArgs)

        store_sol = ApiSolveArgs["store_sol"]
        display_opts = ApiSolveArgs["display_opts"]
        store_hst = ApiSolveArgs["store_hst"]
        hot_start = ApiSolveArgs["hot_start"]

        res = opt_engine(
            opt_prob,
            store_sol=store_sol,
            display_opts=display_opts,
            store_hst=store_hst,
            hot_start=hot_start,
        )

        self.Parameters = res[1]
        self.OFvalue = res[0]

        return res

    def FW1Calibration(self, SpatialVarFun, OptimizationArgs, printError=None):
        """FW1Calibration.

        this function runs the calibration algorithm for the conceptual distributed
        hydrological model

        Parameters
        ----------
        ConceptualModel: [function]
            conceptual model and it should contain a function called simulate

        2-Basic_inputs:
            1-p2: [List]
                list of unoptimized parameters
                p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                p2[1] = catchment area in km2
            2-init_st: [list]
                initial values for the state variables [sp,sm,uz,lz,wc] in mm
            3-UB: [Numeric]
                upper bound of the values of the parameters
            4-LB: [Numeric]
                Lower bound of the values of the parameters
        3-Q_obs: [Numeric]
            Observed values of discharge

        6-lumpedParNo: [int]
            nomber of lumped parameters, you have to enter the value of
            the lumped parameter at the end of the list, default is 0 (no lumped parameters)
        7-lumpedParPos: [List]
            list of order or position of the lumped parameter among all
            the parameters of the lumped model (order starts from 0 to the length
            of the model parameters), default is [] (empty), the following order
            of parameters is used for the lumped HBV model used
            [ltt, utt, rfcf, sfcf, ttm, cfmax, cwh, cfr, fc, beta, e_corr, etf, lp,
            c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
        8-objective_function: [function]
            objective function to calculate the performance of the model
            and to be used in the calibration
        9-*args:
            other arguments needed on the objective function

        Returns
        -------
        st: [4D array]
            state variables
        q_out: [1D array]
            calculated Discharge at the outlet of the catchment
        q_uz: [3D array]
            Distributed discharge for each cell

        Example
        -------
        PrecPath = prec_path="meteodata/4000/calib/prec"
        Evap_Path = evap_path="meteodata/4000/calib/evap"
        TempPath = temp_path="meteodata/4000/calib/temp"
        FlowAccPath = "GIS/4000/acc4000.tif"
        FlowDPath = "GIS/4000/fd4000.tif"
        ParPath = "meteodata/4000/"+"parameters.txt"
        p2=[1, 227.31]
        st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                          FlowAccPath,FlowDPath,ParPath,p2)
        """
        # input dimensions
        # [rows,cols] = self.FlowAcc.ReadAsArray().shape
        # [fd_rows,fd_cols] = self.FlowDirArr.shape
        # assert fd_rows == self.rows and fd_cols == self.cols, "all input data should have the same number of rows"

        # input dimensions
        assert (
            np.shape(self.Prec)[0] == self.rows
            and np.shape(self.ET)[0] == self.rows
            and np.shape(self.Temp)[0] == self.rows
        ), "all input data should have the same number of rows"
        assert (
            np.shape(self.Prec)[1] == self.cols
            and np.shape(self.ET)[1] == self.cols
            and np.shape(self.Temp)[1] == self.cols
        ), "all input data should have the same number of columns"
        assert (
            np.shape(self.Prec)[2] == np.shape(self.ET)[2] and np.shape(self.Temp)[2]
        ), "all meteorological input data should have the same length"

        # basic inputs
        # check if all inputs are included
        # assert all(["p2","init_st","UB","LB","snow "][i] in Basic_inputs.keys() for i in range(4)), "Basic_inputs should contain ['p2','init_st','UB','LB'] "

        ### optimization

        # get arguments
        ApiObjArgs = OptimizationArgs[0]
        pll_type = OptimizationArgs[1]
        ApiSolveArgs = OptimizationArgs[2]
        # check optimization arguement
        assert type(ApiObjArgs) == dict, "store_history should be 0 or 1"
        assert type(ApiSolveArgs) == dict, "history_fname should be of type string "

        print("Calibration starts")

        ### calculate the objective function
        def opt_fun(par):
            try:
                # distribute the parameters
                SpatialVarFun.Function(
                    par
                )  # , kub=SpatialVarFun.Kub, klb=SpatialVarFun.Klb, Maskingum=SpatialVarFun.Maskingum
                self.Parameters = SpatialVarFun.Par3d
                # run the model
                Wrapper.FW1(self)
                # calculate performance of the model
                try:
                    # error = self.OF(self.QGauges, self.qout, self.quz_routed, self.qlz_translated,*[self.GaugesTable])
                    error = self.OF(self.QGauges, self.qout, *[self.GaugesTable])
                except TypeError:  # if no of inputs less than what the function needs
                    assert (
                        False
                    ), "the objective function you have entered needs more inputs please enter then in a list as *args"

                # print error
                if printError != 0:
                    print(round(error, 3))
                    print(par)

                fail = 0
            except:
                error = np.nan
                fail = 1

            return error, [], fail

        ### define the optimization components
        opt_prob = Optimization("HBV Calibration", opt_fun)
        for i in range(len(self.LB)):
            opt_prob.addVar(
                "x{0}".format(i), type="c", lower=self.LB[i], upper=self.UB[i]
            )

        print(opt_prob)

        opt_engine = HSapi(pll_type=pll_type, options=ApiObjArgs)

        store_sol = ApiSolveArgs["store_sol"]
        display_opts = ApiSolveArgs["display_opts"]
        store_hst = ApiSolveArgs["store_hst"]
        hot_start = ApiSolveArgs["hot_start"]

        res = opt_engine(
            opt_prob,
            store_sol=store_sol,
            display_opts=display_opts,
            store_hst=store_hst,
            hot_start=hot_start,
        )

        self.Parameters = res[1]
        self.OFvalue = res[0]

        return res

    def lumpedCalibration(self, Basic_inputs, OptimizationArgs, printError=None):
        """runCalibration.

        this function runs the calibration algorithm for the Lumped conceptual hydrological model

        Parameters
        ----------
        ConceptualModel:
            [function] conceptual model and it should contain a function called simulate
        data:
            [numpy array] meteorological data as array with the first column as precipitation
            second as evapotranspiration, third as temperature and forth column as
            long term average temperature
        Basic_inputs:
            1-p2:
                [List] list of unoptimized parameters
                p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                p2[1] = catchment area in km2
            2-init_st:
                [list] initial values for the state variables [sp,sm,uz,lz,wc] in mm
            3-UB:
                [Numeric] upper bound of the values of the parameters
            4-LB:
                [Numeric] Lower bound of the values of the parameters
        Q_obs:
            [Numeric] Observed values of discharge

        lumpedParNo:
            [int] nomber of lumped parameters, you have to enter the value of
            the lumped parameter at the end of the list, default is 0 (no lumped parameters)
        lumpedParPos:
            [List] list of order or position of the lumped parameter among all
            the parameters of the lumped model (order starts from 0 to the length
            of the model parameters), default is [] (empty), the following order
            of parameters is used for the lumped HBV model used
            [ltt, utt, rfcf, sfcf, ttm, cfmax, cwh, cfr, fc, beta, e_corr, etf, lp,
            c_flux, k, k1, alpha, perc, pcorr, Kmuskingum, Xmuskingum]
        objective_function:
            [function] objective function to calculate the performance of the model
            and to be used in the calibration
        *args:
            other arguments needed on the objective function

        Returns
        -------
        st: [4D array]
            state variables
        q_out: [1D array]
            calculated Discharge at the outlet of the catchment
        q_uz: [3D array]
            Distributed discharge for each cell

        Examples
        --------
            PrecPath = prec_path="meteodata/4000/calib/prec"
            Evap_Path = evap_path="meteodata/4000/calib/evap"
            TempPath = temp_path="meteodata/4000/calib/temp"
            FlowAccPath = "GIS/4000/acc4000.tif"
            FlowDPath = "GIS/4000/fd4000.tif"
            ParPath = "meteodata/4000/"+"parameters.txt"
            p2=[1, 227.31]
            st, q_out, q_uz_routed = RunModel(PrecPath,Evap_Path,TempPath,DemPath,
                                              FlowAccPath,FlowDPath,ParPath,p2)
        """
        # basic inputs
        # check if all inputs are included
        assert all(
            ["Route", "RoutingFn"][i] in Basic_inputs.keys() for i in range(2)
        ), "Basic_inputs should contain ['p2','init_st','UB','LB'] "

        Route = Basic_inputs["Route"]
        RoutingFn = Basic_inputs["RoutingFn"]
        if "InitialValues" in Basic_inputs.keys():
            InitialValues = Basic_inputs["InitialValues"]
        else:
            InitialValues = []

        ### optimization

        # get arguments
        ApiObjArgs = OptimizationArgs[0]
        pll_type = OptimizationArgs[1]
        ApiSolveArgs = OptimizationArgs[2]
        # check optimization arguement
        assert isinstance(ApiObjArgs, dict), "store_history should be 0 or 1"
        assert isinstance(ApiSolveArgs, dict), "history_fname should be of type string "

        # assert history_fname[-4:] == ".txt", "history_fname should be txt file please change extension or add .txt ad the end of the history_fname"

        print("Calibration starts")

        ### calculate the objective function
        def opt_fun(par):
            try:
                # parameters
                self.Parameters = par
                # run the model
                Wrapper.Lumped(self, Route, RoutingFn)
                # calculate performance of the model
                try:
                    error = self.OF(
                        self.QGauges[self.QGauges.columns[-1]], self.Qsim, *self.OFArgs
                    )
                    g = [
                        2 * par[-2] * par[-1] / self.dt,
                        (2 * par[-2] * (1 - par[-1])) / self.dt,
                    ]
                except TypeError:  # if no of inputs less than what the function needs
                    assert (
                        False
                    ), "the objective function you have entered needs more inputs please enter then in a list as *args"

                if printError != 0:
                    print(
                        f"Error = {round(error, 3)} Inequality Const = {np.round(g, 2)}"
                    )
                    # print(par)
                fail = 0
            except:
                error = np.nan
                g = []
                fail = 1
            return error, g, fail

        ### define the optimization components
        opt_prob = Optimization("HBV Calibration", opt_fun)

        if InitialValues != []:
            for i in range(len(self.LB)):
                opt_prob.addVar(
                    "x{0}".format(i),
                    type="c",
                    lower=self.LB[i],
                    upper=self.UB[i],
                    value=InitialValues[i],
                )
        else:
            for i in range(len(self.LB)):
                opt_prob.addVar(
                    "x{0}".format(i), type="c", lower=self.LB[i], upper=self.UB[i]
                )

        opt_prob.addObj("f")

        opt_prob.addCon("g1", "i")
        opt_prob.addCon("g2", "i")
        # print(opt_prob)
        opt_engine = HSapi(pll_type=pll_type, options=ApiObjArgs)

        # parse the ApiSolveArgs inputs
        # availablekeys = ['store_sol',"display_opts","store_hst","hot_start"]
        store_sol = ApiSolveArgs["store_sol"]
        display_opts = ApiSolveArgs["display_opts"]
        store_hst = ApiSolveArgs["store_hst"]
        hot_start = ApiSolveArgs["hot_start"]

        # for i in range(len(availablekeys)):
        # if availablekeys[i] in ApiSolveArgs.keys():
        # exec(availablekeys[i] + "=" + str(ApiSolveArgs[availablekeys[i]]))
        # print(availablekeys[i] + " = " + str(ApiSolveArgs[availablekeys[i]]))

        res = opt_engine(
            opt_prob,
            store_sol=store_sol,
            display_opts=display_opts,
            store_hst=store_hst,
            hot_start=hot_start,
        )

        self.OFvalue = res[0]
        self.Parameters = res[1]

        return res

    def ListAttributes(self):
        """Print Attributes List."""

        print("\n")
        print(
            "Attributes List of: "
            + repr(self.__dict__["name"])
            + " - "
            + self.__class__.__name__
            + " Instance\n"
        )
        self_keys = list(self.__dict__.keys())
        self_keys.sort()
        for key in self_keys:
            if key != "name":
                print(str(key) + " : " + repr(self.__dict__[key]))

        print("\n")


if __name__ == "__main__":
    print("Calibration")



"""Created on Wed Jun 27 19:17:15 2018.

@author: Mostafa
"""

import numpy as np
from pyramids.raster import Raster as raster

from Hapi.rrm.routing import Routing as routing


class DistributedRRM:
    """DistributedRRM class runs simulation in lumped form for each cell separetly and rout the discharge between the cells following the rivernetwork.

    Methods:
        1-RunLumpedRRM
        2-SpatialRouting
        3-DistMaxBas1
        4-DistMaxBas2
        5-Dist_HBV2
    """

    def __init__(self):
        pass

    @staticmethod
    def RunLumpedRRM(Model):
        """RunLumpedRRM method runs the rainfall runoff lumped model (HBV, GR4,...) separately for each cell and return a time series of arrays.

        Inputs:
        ----------
            1-ConceptualModel:
                [function] conceptual model function
            2-Raster:
                [gdal.dataset] raster to get the spatial information (nodata cells)
                raster input could be dem, flow accumulation or flow direction raster of the catchment
                but the nodata value stored in the raster should be far from the
                range of values that could result from the calculation
            3-sp_prec:
                [numpy array] 3d array of the precipitation data, sp_prec should
                have the same 2d dimension of raster input
            4-sp_et:
                [numpy array] 3d array of the evapotranspiration data, sp_et should
                have the same 2d dimension of raster input
            5-sp_temp:
                [numpy array] 3d array of the temperature data, sp_temp should
                have the same 2d dimension of raster input
            6-sp_pars:
                [numpy array] number of 2d arrays of the catchment properties spatially
                distributed in 2d and the third dimension is the number of parameters,
                sp_pars should have the same 2d dimension of raster input
            7-p2:
                [List] list of unoptimized parameters
                p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                p2[1] = catchment area in km2
            8-init_st:
                [list] initial state variables values [sp, sm, uz, lz, wc]. default=None
            9-ll_temp:
                [numpy array] 3d array of the long term average temperature data
            10-q_init:
                [float] initial discharge m3/s
        Outputs:
        ----------
            1-statevariables : [numpy ndarray]
                 4D array (rows,cols,time,states) states are [sp,wc,sm,uz,lv]
            2-qlz : [numpy ndarray]
                3D array of the lower zone discharge (lumped calculation for each cell separately)
            3-quz : [numpy ndarray]
                3D array of the upper zone discharge
        """
        Model.statevariables = np.zeros(
            [Model.rows, Model.cols, Model.TS, 5], dtype=np.float32
        )
        # Model.statevariables[:] = np.nan
        Model.quz = np.zeros([Model.rows, Model.cols, Model.TS], dtype=np.float32)
        # Model.q_uz[:] = np.nan
        Model.qlz = np.zeros([Model.rows, Model.cols, Model.TS], dtype=np.float32)

        for x in range(Model.rows):
            for y in range(Model.cols):
                # only for cells in the domain
                if not np.isnan(Model.FlowAccArr[x, y]):
                    (
                        Model.quz[x, y, :],
                        Model.qlz[x, y, :],
                        Model.statevariables[x, y, :, :],
                    ) = Model.LumpedModel.Simulate(
                        prec=Model.Prec[x, y, :],
                        temp=Model.Temp[x, y, :],
                        et=Model.ET[x, y, :],
                        ll_temp=Model.ll_temp[x, y, :],
                        par=Model.Parameters[x, y, :],
                        init_st=Model.InitialCond,
                        q_init=Model.q_init,
                        snow=Model.Snow,
                    )

        area_coef = Model.CatArea / Model.px_tot_area
        # convert quz from mm/time step to m3/sec
        Model.quz = (
            Model.quz * Model.px_area * area_coef / Model.conversionfactor
        )  # Timef*3.6
        # convert Qlz to m3/sec
        Model.qlz = (
            Model.qlz * Model.px_area * area_coef / Model.conversionfactor
        )  # Timef*3.6

    @staticmethod
    def SpatialRouting(Model):
        """SpatialRouting method routes the discharge from cell to another following the flow direction input raster.

        Inputs:
        ----------
            1-qlz:
                [numpy ndarray] 3D array of the lower zone discharge
            2-quz:
                [numpy ndarray] 3D array of the upper zone discharge
            3-flow_acc:
                [gdal.dataset] flow accumulation raster file of the catchment (clipped to the catchment only)
            4-flow_direct:
                [gdal.dataset] flow Direction raster file of the catchment (clipped to the catchment only)
            5-sp_pars:
                [numpy ndarray] 3D array of the parameters
            6-p2:
                [List] list of unoptimized parameters
                p2[0] = tfac, 1 for hourly, 0.25 for 15 min time step and 24 for daily time step
                p2[1] = catchment area in km2

        Outputs:
        ----------
            1-qout:
                [numpy array] 1D timeseries of discharge at the outlet of the catchment
                of unit m3/sec
            2-quz_routed:
                [numpy ndarray] 3D array of the upper zone discharge  accumulated and
                routed at each time step
            3-qlz_translated:
                [numpy ndarray] 3D array of the lower zone discharge translated at each time step
        """
        #    # routing lake discharge with DS cell k & x and adding to cell Q
        #    q_lake=Routing.Muskingum_V(q_lake,q_lake[0],sp_pars[lakecell[0],lakecell[1],10],sp_pars[lakecell[0],lakecell[1],11],p2[0])
        #    q_lake=np.append(q_lake,q_lake[-1])
        #    # both lake & Quz are in m3/s
        #    #new
        #    quz[lakecell[0],lakecell[1],:]=quz[lakecell[0],lakecell[1],:]+q_lake

        ### cells at the divider
        Model.quz_routed = np.zeros_like(Model.quz)
        """
        lower zone discharge is going to be just translated without any attenuation
        in order to be able to calculate total discharge (uz+lz) at internal points
        in the catchment
        """
        Model.qlz_translated = np.zeros_like(Model.quz)
        # Model.Qtot = np.zeros_like(Model.quz)
        # for all cell with 0 flow acc put the quz
        for x in range(Model.rows):  # no of rows
            for y in range(Model.cols):  # no of columns
                if not np.isnan(Model.FlowAccArr[x, y]) and Model.FlowAccArr[x, y] == 0:
                    Model.quz_routed[x, y, :] = Model.quz[x, y, :]
                    Model.qlz_translated[x, y, :] = Model.qlz[x, y, :]

        ### remaining cells
        for j in range(1, len(Model.acc_val)):
            # TODO parallelize
            # all cells with the same acc_val can run at the same time
            for x in range(Model.rows):  # no of rows
                for y in range(Model.cols):  # no of columns
                    # check from total flow accumulation
                    if (
                        not np.isnan(Model.FlowAccArr[x, y])
                        and Model.FlowAccArr[x, y] == Model.acc_val[j]
                    ):
                        if (
                            Model.RouteRiver != "Muskingum"
                            and Model.BankfullDepth[x, y] > 0
                        ):
                            continue
                        else:
                            # for UZ
                            q_uzi = np.zeros(Model.TS)
                            # for lz
                            qlzi = np.zeros(Model.TS)
                            # iterate to route uz and translate lz
                            for i in range(
                                len(Model.FDT[str(x) + "," + str(y)])
                            ):  # Model.acc_val[j]
                                # bring the indexes of the us cell
                                x_ind = Model.FDT[str(x) + "," + str(y)][i][0]
                                y_ind = Model.FDT[str(x) + "," + str(y)][i][1]
                                # sum the Q of the US cells (already routed for its cell)
                                # route first with there own k & xthen sum
                                q_uzi = q_uzi + routing.Muskingum_V(
                                    Model.quz_routed[x_ind, y_ind, :],
                                    Model.quz_routed[x_ind, y_ind, 0],
                                    Model.Parameters[x_ind, y_ind, 10],
                                    Model.Parameters[x_ind, y_ind, 11],
                                    Model.dt,
                                )

                                qlzi = qlzi + Model.qlz_translated[x_ind, y_ind, :]

                            # add the routed upstream flows to the current Quz in the cell
                            Model.quz_routed[x, y, :] = Model.quz[x, y, :] + q_uzi
                            Model.qlz_translated[x, y, :] = Model.qlz[x, y, :] + qlzi
        Model.Qtot = Model.qlz_translated + Model.quz_routed

    @staticmethod
    def DistMaxbas1(Model):
        """DistMaxbas1 method rout the discharge directly to the outlet from each cell using triangular function.

        Parameters
        ----------
        Model : TYPE
            DESCRIPTION.

        Returns
        -------
        None.
        """

        Maxbas = Model.Parameters[:, :, -1]

        for x in range(Model.rows):
            for y in range(Model.cols):
                if not np.isnan(Model.FlowAccArr[x, y]):
                    Model.quz[x, y, :] = routing.TriangularRouting1(
                        Model.quz[x, y, :], Maxbas[x, y]
                    )

    @staticmethod
    def DistMaxbas2(Model):
        """DistMaxbas2 method rout the discharge directly to the outlet from each cell using triangular function, the maxbas parameters are going to be calculated based on the flow path length input.

        Parameters
        ----------
        Model : TYPE
            DESCRIPTION.

        Returns
        -------
        None.
        """

        MAXBAS = np.nanmax(Model.Parameters[:, :, -1])
        # replace novalue cells by nan
        Model.FPLArr[Model.FPLArr == Model.NoDataValue] = np.nan

        MaxFPL = np.nanmax(Model.FPLArr)
        MinFPL = np.nanmin(Model.FPLArr)
        # resize_fun = lambda x: np.round(((((x - min_dist)/(max_dist - min_dist))*(1*maxbas - 1)) + 1), 0)
        resize_fun = lambda g: (
            (((g - MinFPL) / (MaxFPL - MinFPL)) * (1 * MAXBAS - 1)) + 1
        )

        NormalizedFPL = resize_fun(Model.FPLArr)

        for x in range(Model.rows):
            for y in range(Model.cols):
                if not np.isnan(Model.FPLArr[x, y]):
                    Model.quz[x, y, :] = routing.TriangularRouting2(
                        Model.quz[x, y, :], NormalizedFPL[x, y]
                    )

    @staticmethod
    def Dist_HBV2(
        ConceptualModel,
        lakecell,
        q_lake,
        DEM,
        flow_acc,
        flow_acc_plan,
        sp_prec,
        sp_et,
        sp_temp,
        sp_pars,
        p2,
        init_st=None,
        ll_temp=None,
        q_0=None,
    ):
        """original function."""

        n_steps = sp_prec.shape[2] + 1  # no of time steps =length of time series +1
        # intiialise vector of nans to fill states
        dummy_states = np.empty([n_steps, 5])  # [sp,sm,uz,lz,wc]
        dummy_states[:] = np.nan

        # Get the mask
        # mask, no_val = raster.get_mask(DEM)
        mask, no_val = raster.getRasterData(DEM)
        # shape of the fpl raster (rows, columns)-------------- rows are x and columns are y
        x_ext, y_ext = mask.shape
        #    y_ext, x_ext = mask.shape # shape of the fpl raster (rows, columns)------------ should change rows are y and columns are x

        # Get deltas of pixel
        geo_trans = (
            DEM.GetGeoTransform()
        )  # get the coordinates of the top left corner and cell size [x,dx,y,dy]
        dx = np.abs(geo_trans[1]) / 1000.0  # dx in Km
        dy = np.abs(geo_trans[-1]) / 1000.0  # dy in Km
        px_area = dx * dy  # area of the cell

        # Enumerate the total number of pixels in the catchment
        tot_elem = np.sum(
            np.sum([[1 for elem in mask_i if elem != no_val] for mask_i in mask])
        )  # get row by row and search [mask_i for mask_i in mask]

        # total pixel area
        px_tot_area = tot_elem * px_area  # total area of pixels

        # Get number of non-value data

        st = []  # Spatially distributed states
        qlz = []
        quz = []
        # ------------------------------------------------------------------------------
        for x in range(x_ext):  # no of rows
            st_i = []
            q_lzi = []
            q_uzi = []
            #        q_out_i = []
            # run all cells in one row ----------------------------------------------------
            for y in range(y_ext):  # no of columns
                if mask[x, y] != no_val:  # only for cells in the domain
                    # Calculate the states per cell
                    # TODO optimise for multiprocessing these loops
                    #                _, _st, _uzg, _lzg = ConceptualModel.simulate_new_model(avg_prec = sp_prec[x, y,:],
                    _, _st, _uzg, _lzg = ConceptualModel.Simulate(
                        prec=sp_prec[x, y, :],
                        temp=sp_temp[x, y, :],
                        et=sp_et[x, y, :],
                        par=sp_pars[x, y, :],
                        p2=p2,
                        init_st=init_st,
                        ll_temp=None,
                        q_0=q_0,
                        snow=0,
                    )  # extra_out = True
                    # append column after column in the same row -----------------
                    st_i.append(np.array(_st))
                    # calculate upper zone Q = K1*(LZ_int_1)
                    q_lz_temp = np.array(sp_pars[x, y, 6]) * _lzg
                    q_lzi.append(q_lz_temp)
                    # calculate lower zone Q = k*(UZ_int_3)**(1+alpha)
                    q_uz_temp = np.array(sp_pars[x, y, 5]) * (
                        np.power(_uzg, (1.0 + sp_pars[x, y, 7]))
                    )
                    q_uzi.append(q_uz_temp)

                    # print("total = "+str(fff)+"/"+str(tot_elem)+" cell, row= "+str(x+1)+" column= "+str(y+1) )
                else:  # if the cell is novalue-------------------------------------
                    # Fill the empty cells with a nan vector
                    st_i.append(
                        dummy_states
                    )  # fill all states(5 states) for all time steps = nan
                    q_lzi.append(
                        dummy_states[:, 0]
                    )  # q lower zone =nan  for all time steps = nan
                    q_uzi.append(
                        dummy_states[:, 0]
                    )  # q upper zone =nan  for all time steps = nan

            # store row by row-------- ----------------------------------------------------
            # st.append(st_i) # state variables
            st.append(st_i)  # state variables
            qlz.append(np.array(q_lzi))  # lower zone discharge mm/timestep
            quz.append(np.array(q_uzi))  # upper zone routed discharge mm/timestep
            # ------------------------------------------------------------------------------
            # convert to arrays
        st = np.array(st)
        qlz = np.array(qlz)
        quz = np.array(quz)
        # %% convert quz from mm/time step to m3/sec
        area_coef = p2[1] / px_tot_area
        quz = quz * px_area * area_coef / (p2[0] * 3.6)

        no_cells = list(
            set(
                [
                    flow_acc_plan[i, j]
                    for i in range(x_ext)
                    for j in range(y_ext)
                    if not np.isnan(flow_acc_plan[i, j])
                ]
            )
        )
        #    no_cells=list(set([int(flow_acc_plan[i,j]) for i in range(x_ext) for j in range(y_ext) if flow_acc_plan[i,j] != no_val]))
        no_cells.sort()

        # %% routing lake discharge with DS cell k & x and adding to cell Q
        q_lake = routing.Muskingum_V(
            q_lake,
            q_lake[0],
            sp_pars[lakecell[0], lakecell[1], 10],
            sp_pars[lakecell[0], lakecell[1], 11],
            p2[0],
        )
        q_lake = np.append(q_lake, q_lake[-1])
        # both lake & Quz are in m3/s
        # new
        quz[lakecell[0], lakecell[1], :] = quz[lakecell[0], lakecell[1], :] + q_lake
        # %% cells at the divider
        quz_routed = np.zeros_like(quz) * np.nan
        # for all cell with 0 flow acc put the quz
        for x in range(x_ext):  # no of rows
            for y in range(y_ext):  # no of columns
                if mask[x, y] != no_val and flow_acc_plan[x, y] == 0:
                    quz_routed[x, y, :] = quz[x, y, :]
        # %% new
        for j in range(1, len(no_cells)):  # 2):#
            for x in range(x_ext):  # no of rows
                for y in range(y_ext):  # no of columns
                    # check from total flow accumulation
                    if mask[x, y] != no_val and flow_acc_plan[x, y] == no_cells[j]:
                        #                        print(no_cells[j])
                        q_r = np.zeros(n_steps)
                        for i in range(
                            len(flow_acc[str(x) + "," + str(y)])
                        ):  # no_cells[j]
                            # bring the indexes of the us cell
                            x_ind = flow_acc[str(x) + "," + str(y)][i][0]
                            y_ind = flow_acc[str(x) + "," + str(y)][i][1]
                            # sum the Q of the US cells (already routed for its cell)
                            # route first with there own k & xthen sum
                            q_r = q_r + routing.Muskingum_V(
                                quz_routed[x_ind, y_ind, :],
                                quz_routed[x_ind, y_ind, 0],
                                sp_pars[x_ind, y_ind, 10],
                                sp_pars[x_ind, y_ind, 11],
                                p2[0],
                            )
                        #                        q=q_r
                        # add the routed upstream flows to the current Quz in the cell
                        quz_routed[x, y, :] = quz[x, y, :] + q_r
        # %% check if the max flow _acc is at the outlet
        #    if tot_elem != np.nanmax(flow_acc_plan):
        #        raise ("flow accumulation plan is not correct")
        # outlet is the cell that has the max flow_acc
        outlet = np.where(
            flow_acc_plan == np.nanmax(flow_acc_plan)
        )  # np.nanmax(flow_acc_plan)
        outletx = outlet[0][0]
        outlety = outlet[1][0]
        # %%
        qlz = np.array(
            [np.nanmean(qlz[:, :, i]) for i in range(n_steps)]
        )  # average of all cells (not routed mm/timestep)
        # convert Qlz to m3/sec
        qlz = qlz * p2[1] / (p2[0] * 3.6)  # generation

        qout = qlz + quz_routed[outletx, outlety, :]

        return qout, st, quz_routed, qlz, quz
